#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include <omp.h>

/* arbitrarily chosen four million */
#define N ((int) 4e6)

/* arbitrary function */
static double foo (double x) {
  return cos(sin(x));
}

int main (void) {
  int i;
  double * a, * b;

  a = malloc(N * sizeof(double));
  b = malloc(N * sizeof(double));

  if (a == NULL || b == NULL) {
    perror("malloc");
    exit(EXIT_FAILURE);	    
  }

  /* fill arrays with arbitrary values */
#pragma omp parallel for
  for (i = 0; i < N; i++) {
    a[i] = 2.0*i;
    b[i] = a[i] - 1.0;
  }

  /* a */
  {
    double x = N;

#pragma omp parallel for
    for (i = 0; i < (int) sqrt(x); i++) {
      a[i] = 2.3*i;

      if (i < 10)
	b[i] = a[i];
    }
  }

  /* b */
  {
    int n = N;
    int flag;

    /*
     * Can't trivially parallelize
     * for loop control expression
     * not in 'canonical' form.
     */
    flag = 0;
    for (i = 0; (i < n) && (!flag); i++) {
      a[i] = 2.3 * i;

      if (a[i] < b[i])
	flag = 1;
    }
  }

  /* c */
  {
    int n = N;

    /*
     * Trivially parallelized, as long
     * as foo has no hidden side-effects.
     */
#pragma omp parallel for
    for (i = 0; i < n; i++)
      a[i] = foo(i);
  }

  /* d */
  {
    int n = N;

    /*
     * Again, everything is fine as
     * long as foo has no side-effects.
     */
#pragma omp parallel for
    for (i = 0; i < n; i++) {
      a[i] = foo(i);

      if (a[i] < b[i])
	a[i] = b[i];
    }
  }

  /* e */
  {
    int n = N;

    /*
     * Can't do parallelization,
     * break statement within
     * loop.
     */
    for (i = 0; i < n; i++) {
      a[i] = foo(i);

      if (a[i] < b[i])
	break;
    }
  }

  /* f */
  {
    int n = N;
    double dotp;

    dotp = 0;
#pragma omp parallel for reduction(+:dotp)
    for (i = 0; i < n; i++)
      dotp += a[i]*b[i];
  }

  /* g */
  {
    int k = N/2;

    /*
     * This looks shoddy but works
     * since there is no overlap in
     * the data dependency.
     */
#pragma omp parallel for
    for (i = k; i < 2*k; i++)
      a[i] = a[i] + a[i-k];
  }

  /* h */
  {
    int n = N;
    int k = N/2-1;
    double bb = 0.23; /* arbitrary */

    /*
     * If n >= 2*k we have data
     * dependency issue, data
     * race between threads.
     */
#pragma omp parallel for if (n < 2*k)
    for (i = k; i < n; i++)
      a[i] = bb * a[i-k];
  }

  free(b);
  free(a);

  exit(EXIT_SUCCESS);
}
